/************************* * * * * * * * * * * * * ***************************
    Copyright (c) 1999-2003 Ryan Bobko
                       ryan@ostrich-emulators.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.     
************************** * * * * * * * * * * * * **************************/

#ifndef PLUGINS_H
#define PLUGINS_H

#include <qstring.h>

#include "tableops.h"
#include "resultset.h"
#include "qhaccutils.h"

class QHacc;

class QHaccPlugin {
 public:
	virtual QString understands() const =0;
	virtual QString description() const =0;
};

// we'll use these typedefs when calling the libraries
typedef QHaccPlugin * (* Creator)();
typedef void (* Destroyer)( QHaccPlugin * ) ;


/****************************************************************/
/**                                                            **/
/** QHACCIOPLUGIN                                              **/
/**   A class with some amount of I/O capability. This is a    **/
/**   baseclass for importers, exporters, and also QHACC_HOME  **/
/**   readers.                                                 **/
/**                                                            **/
/**   An importer should at least implement load() and exprt() **/
/**   An exporter should at least implement imprt() and save() **/
/**                                                            **/
/****************************************************************/

class QHaccIOPlugin : public QHaccPlugin {
 public:
	/**
	 * Connect to the given QHACC_HOME, but don't load any data.
	 * @param eng The QHacc engine, should it be needed
	 * @param home The location of the datafile(s)
	 * @param error Any error that is raised by the connection
	 * @return True, if the load worked
	 */
	virtual bool connect( QHacc * eng, const QString& home, QString& error ) =0;
	/**
	 * Load data from the given home into the plugin's datastore. Importers
	 * must overload this function meaningfully.
	 * @param engine The QHacc engine, should it be needed
	 * @param home The location of the datafile(s)
	 * @param error Any error that is raised by the load
	 * @return True, if the load worked
	 **/
	virtual bool load( QHacc * engine, const QString& home, QString& error ) =0;

	/**
	 * Furnish the internal data to other objects. Importers need this function.
	 * @param rs A set of six datasets correlating to the LocalFiles' structure
	 * @return True, in all cases
	 */
	virtual bool exprt( QHaccResultSet * rs ) =0;
	
	/**
	 * Load the data furnished by some other exprt() function. 
	 * @param rs The data to load
	 * @return True, in all cases
	 **/
	virtual bool imprt( QHaccResultSet * rs ) =0;

	/**
	 * Save data
	 * @param home The QHACC_HOME in which to save
	 * @param protect Whether or not to protect the data (plugin dependent)
	 * @param error Any error that is raised by the save
	 * @return True, if the save works
	 **/
	virtual bool save( const QString& home, bool protect, QString& error ) =0;
};

/****************************************************************/
/** QHACCDBPLUGIN                                              **/
/**   Finally, a class that can load and save, and also        **/
/**   interact with a database!                                **/
/**                                                            **/
/****************************************************************/
class QHaccDBPlugin : public QHaccIOPlugin {
 public:
	virtual bool load( QHacc * engine, const QString& home, QString& error ) =0;
	virtual bool load( Table, QHaccResultSet& ) =0;

	// create commands to make a new QHACC_HOME
	/**
	 * Create a creation script for the given QHACC_HOME
	 * @param home The QHACC_HOME to be created
	 * @return The creation script
	 **/
	virtual QString create( const QString& home ) const =0;

	/**
	 * @return A description of the plugin
	 **/
	virtual QString description() const =0; // user description of this plugin
	/**
	 * @return What QHACC_HOME identifier this plugin understands
	 **/
	virtual QString understands() const =0; // what HOME keyword do I know?

	virtual int add( Table, const TableRow& ) =0;

	virtual int updateWhere( Table, const TableSelect&, const TableUpdate& ) =0;
	virtual int updateWhere( Table, const TableSelect&, const TableRow& ) =0;

	virtual bool getWhere( Table, TableSelect *, uint size,
												 QHaccResultSet *& results ) =0;
	virtual bool getWhere( Table, const TableGet&, TableSelect *, uint size,
												 QHaccResultSet *& results ) =0;
	virtual bool getWhere( Table, const TableSelect&,
												 QHaccResultSet *& results ) =0;

	virtual uint deleteWhere( Table, const TableSelect& ) =0;
	virtual uint deleteWhere( Table, QHaccResultSet * ) =0;

	virtual TableCol max( Table, int col ) =0;
	virtual TableCol min( Table, int col ) =0;

	/**
	 * How many rows are in the given table?
	 * @param t The table in question
	 * @returns The number of rows in the given table
	 **/
	virtual uint cnt( Table t ) =0;  // rows in the given table

	/**
	 * @returns True, if the datastore needs saving
	 **/
	virtual bool dirty() const =0; // data needs to be saved?

	/**
	 * Prepare the database for a large data load
	 * @param t The table that will be loaded
	 * @param num The number of rows that will be added
	 **/
	virtual void startLoad( Table t, uint num =0 ) =0;

	/**
	 * Bring the given table out of loading setup
	 * @param t The table
	 **/
	virtual void stopLoad( Table t ) =0;
};

#endif
