/************************* * * * * * * * * * * * * ***************************
    Copyright (c) 1999-2003 Ryan Bobko
                       ryan@ostrich-emulators.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.     
************************** * * * * * * * * * * * * **************************/

#ifndef _QHACC_H
#define _QHACC_H

#include "tableops.h"
#include "resultset.h"

#include <qfont.h>
#include <qcolor.h>
#include <qobject.h>
#include <qstring.h>
#include <qdatetime.h>

typedef TableRow Ledger;
typedef TableRow Account;
typedef TableRow Transaction;

class QLibrary;
class QHaccPlugin;
class QHaccDBPlugin;
class QHaccIOPlugin;
class PluginManager;

/**
 * @short QHacc personal financial software database engine
 * @author Ryan Bobko <ryan@ostrich-emulators.com>
 */
class QHacc : public QObject {
  Q_OBJECT
		public:
	friend class QHaccExt;
	static const int ROLLOVER, PIDATABASE, PIIMPORTER, PIEXPORTER, PITYPES;

	/**
	 * Creates an instance and finds all available plugins starting 
	 * in the given directory
	 *
	 * @param root Root of plugin-finding search
	 */
	QHacc( const char * root );
	
	/**
	 * Deletes the currently-loaded database plugin and closes 
	 * the library, if open
	 */
  ~QHacc();
	
	/**
	 * @return The current QHACC_HOME variable
	 */
	QString getHome() const;
	
	/**
	 * Sets the current QHACC_HOME variable and loads all the data therein
	 *
	 * @param QHACC_HOME A String representing the new datastore location
	 */
	bool setHome( const QString& QHACC_HOME );

	/**
	 * @return True, if the current QHACC_HOME is contained in 
	 * QHacc's native file structure
	 */
	bool homeIsLocalFiles() const;
	
	/**
	 * Save the current state of QHACC_HOME to the database
	 * @param error A description of any error thrown by the save
	 * @return True if the save was successful
	 */
	bool save( QString& error );

	/**
	 * Create an export of the database into an array of ResultSets
	 * @param rs The pre-sized array of ResultSets
	 */
	void exprt( QHaccResultSet * rs );

	/**
	 * Import the array of ResultSets into the database. This function allows
	 * direct database access, so it should be used with extreme care. 
	 * @param rs The pre-sized array of ResultSets
	 */
	void imprt( QHaccResultSet * rs );
	
	/**
	 * Create the given account
	 * @param a The account to be added
	 */
	void addA( const Account& a );

	/**
	 * Update the Accounts table to make the account that looks like @p old look
	 * like @p newer
	 * @param old The account to be modified
	 * @param newer The new account data
	 */
	void updateA( const Account& old, const Account& newer );

	/**
	 * Remove the given account, all its transactions, and all its transactions'
	 * splits. It recalculates all account balances affected by the deletions.
	 * @param a The account to be removed
	 */
	void removeA( const Account& a );

	/**
	 * Returns the given account. The identifier can be the account's full name,
	 * its short name, or its id. Matches are checked in that order.
	 * @param str The account identifier.
	 * @return The account specified by the given string.
	 */
	Account getA( const QString& str );

	/**
	 * Returns the given account
	 * @param aid The account ID
	 * @return The account specified by the ID
	 */
	Account getA( uint aid );


	/**
	 * Get the full name of the given account. The full name is the account's
	 * name prefixed by its parent's account name. Top-level accounts have the
	 * same short name and full name.
	 * @param a The account in question
	 * @return The account's full name
	 */
	QString getFNameOfA( const Account& a );

	/**
	 * Get the full name of the given account. The full name is the account's
	 * name prefixed by its parent's account name. Top-level accounts have the
	 * same short name and full name.
	 * @param id The id of the account in question
	 * @return The account's full name
	 */
	QString getFNameOfA( uint id );

	/**
	 * Get all accounts
	 * @param excludeCats Don't include accounts whose category flag is true.
	 * @param tg Which columns to retrieve
	 * @param rs A dynamically-created ResultSet containing the desired accounts.
	 * (This must be deleted by the caller.)
	 */
	void getAs( bool excludeCats, const TableGet& tg, QHaccResultSet *& rs );

	/**
	 * Get a default account to be filled in later. This creates a row suitable
	 * for insertion into the Accounts table, but does not insert it. The row
	 * can be modified for later insertion.
	 *
	 * @return An account with default values set.
	 */
	Account getBlankA() const;

	/**
	 * Get the name of the type of the specified account 
	 * @param a The account in question
	 * @return The name of the account type of the give account
	 */
	QString getTypeNameOfA( const Account& a );

	/**
	 * Get the balance of the given account on the given day. This is the balance
	 * before any transactions have been entered for that day.
	 * @param a The account in question
	 * @param d The date
	 * @return The balance of the account as of that date
	 */
	int getABalOn( const Account& a, const QDate& d );

	/**
	 * Get the reconciled balance of the given account on the given day. This is
	 * the reconciled balance before any transactions have been entered for that
	 * day.
	 * @param a The account in question
	 * @param d The date
	 * @return The reconciled balance of the account as of that date
	 */
	int getARBalTo( const Account&, const QDate& );

	/**
	 * Determine if the given acccount increases on the left side of the
	 * T-balance or not
	 * @param a The account in question
	 * @return True, if the account is a Left-Plus Account
	 */
	bool isLPA( const Account& a );

	/**
	 * Add the given set of rows as one splitgroup
	 * @param rs The set of transactions to link into one splitgroup
	 * @return The new splitgroup id of the rows
	 */
	uint addTSG( QHaccResultSet * rs );

	/**
	 * Delete the transactions from the given splitgroup. Recalculate the
	 * affected account balances.
	 */
	void removeTSG( const TableCol& sg );

	/**
	 * Update the Transactions table to make the splitgroup specified by @p sg 
	 * contain the rows given in @p newsplits.
	 * affected account balances.
	 * @param sg The splitgroup id that needs changing
	 * @param newsplits The new splits to associate into the splitgroup
	 */
	void updateTSG( const TableCol& sg, QHaccResultSet * newsplits );

	/**
	 * Set the reconcile flag on the given transaction to the given state. Note
	 * that this  function does not recalculate the account balance of the
	 * transaction.
	 * @param t The transaction to change
	 * @param recmode the new reconcile mode
	 */
	void setTRecNR( const Transaction& t, uint recmode );

	/**
	 * Set the reconcile flag on the given transactions to the given state. This
	 * function will recalculate account balances as needed.
	 * @param t The set of transactions to change
	 * @param recmode the new reconcile mode
	 */	
	void setTRec( QHaccTable * t, uint recmode );

	/**
	 * Get the transaction identified by the given id
	 * @param tid The transaction ID to return
	 * @return The specified transaction
	 */
	Transaction getT( uint tid );
	
	/**
	 * Overloaded version of the getWhere function that only queries 
	 * the transaction table.
	 * @param ts An array of TableSelect objects
	 * @param tssz The size of the ts array
	 * @param rs A dynamically-created ResultSet with the results of the query
	 * (this has to be deleted by the calling object.)
	 */
	void getTWhere( TableSelect * ts, uint tssz, QHaccResultSet *& rs );

	/**
	 * Overloaded version of the getWhere function that only queries 
	 * the transaction table for the given account's transactions.
	 * @param a The account whose transactions should be returned
	 * @param tg A list of columns to retrieve
	 * @param rs A dynamically-created ResultSet with the results of the query
	 * (this has to be deleted by the calling object.)
	 */
	void getTForA( const Account& a, const TableGet& tg, QHaccResultSet *& rs );

	/**
	 * Get transactions in the same splitgroup as the given transaction. Note: 
	 * the resulting set of transactions does not include the transaction 
	 * passed as @p t.
	 * @param t The transaction
	 * @param rs A dynamically-created ResultSet with the results of the query
	 * (this has to be deleted by the calling object.)
	 */		
	void getTSplits( const Transaction& t, QHaccResultSet *& rs );

	/**
	 * Get transactions in the same splitgroup as the given transaction. Note: 
	 * the resulting set of transactions does not include include the transaction
	 * whose ID is specified by @p id.
	 * @param t The transaction
	 * @param rs A dynamically-created ResultSet with the results of the query
	 * (this has to be deleted by the calling object.)
	 */
	void getTSplits( uint tid, QHaccResultSet *& rs );

	/**
	 * Get a default transaction for the given account and ledger. This creates
	 * a row suitable for insertion into the Transactions table, but does not
	 * insert it. The row can be modified for later insertion.
	 * @param a The parent account of the transaction
	 * @param l The parent ledger of the transaction
	 * @return A transaction with default values set.
	 */
	Transaction getBlankTForAL( const Account& a, const Ledger& l ) const;

	/**
	 * Get the memorized transactions for the given account.
	 * @param a The account in question
	 * @param rs A dynamically-created ResultSet with the results of the query
	 * (this has to be deleted by the calling object.)
	 */
	void getMSubsetForA( const Account& a, QHaccResultSet *& rs );
	/**
	 * Add/remove the given transaction to the list of memorized transactions.
	 * Memorized transactions are account-specific.
	 * @param tid The ID of the transaction to be memorized
	 * @param add If true, add the transaction to the memorized list, else
	 * remove it.
	 */
	void memorize( uint tid, bool add );

	/**
	 * Add the specified transaction to the given ledger. Update its date to
	 * @p d.
	 * @param tid The ID of the transaction to be added.
	 * @param d The date of the transaction upon insertion.
	 * @param l The ledger to which the transaction should be added.
	 */
	void addMtoL( uint tid, QDate d, const Ledger& l );
	

	/**
	 * Add the specified ledger to the Ledgers table.
	 * @param l The ledger to be added.
	 */
	void addL( const Ledger& l );

	/**
	 * Update the Ledgers table to make the 
	 * Update the Ledgers table to make the ledger that looks like @p old look
	 * like @p newer
	 * @param old The ledger to be modified
	 * @param newer The new ledger data.
	 */
	void updateL( const Ledger& old, const Ledger& newer );

	/**
	 * Remove the given ledger, all its transactions, and all its transactions'
	 * splits. It recalculates all account balances affected by the deletions.
	 * @param l The ledger to be removed
	 */
	void removeL( const Ledger& l );

	/**
	 * Get all ledgers
	 * @param rs A dynamically-created ResultSet containing the desired accounts.
	 * (This must be deleted by the caller.)
	 */
	void getLs( QHaccResultSet *& rs );
	
	/**
	 * Get the given ledger.
	 * @param str The ledger's name
	 * @return The ledger specified by the given string.
	 */
	Ledger getL( const QString& str );
	
	/**
	 * Get the given ledger
	 * @param lid The ID of the ledger to be returned
	 * @return The ledger specified
	 */
	Ledger getL( uint lid );

	/**
	 * Get a default ledger to be filled in later. This creates a row suitable
	 * for insertion into the Ledgers table, but does not insert it. The row
	 * can be modified for later insertion.
	 *
	 * @return A ledger with default values set.
	 */
	Ledger getBlankL();


	/**
	 * Get the specified preference
	 * @return the preference value as a string
	 */
	QString getSP( const QString& ) const;

	/**
	 * Get the specified preference
	 * @return the preference value as a color
	 */
	QColor getCP( const QString& ) const;

	/**
	 * Get the specified preference
	 * @return the preference value as a date
	 */
	QDate getDP( const QString& ) const;

	/**
	 * Get the specified preference
	 * @return the preference value as a font
	 */
	QFont getWP( const QString& ) const;

	/**
	 * Get the specified preference
	 * @return the preference value as a floating-point value
	 */
	float getFP( const QString& ) const;

	/**
	 * Get the specified preference
	 * @return the preference value as a boolean value
	 */
	bool getBP( const QString& ) const;

	/**
	 * Get the specified preference
	 * @return the preference value as an integer
	 */
	int getIP( const QString& ) const;

	/**
	 * Set the specified preference @p pref to the given value @p val.
	 * @param pref The preference in question
	 * @param val The new preference value
	 */
	void setSP( const QString& pref, QString val );

	/**
	 * Set the specified preference @p pref to the given value @p val.
	 * @param pref The preference in question
	 * @param val The new preference value
	 */
	void setCP( const QString& pref, QColor val );

	/**
	 * Set the specified preference @p pref to the given value @p val.
	 * @param pref The preference in question
	 * @param val The new preference value
	 */
	void setDP( const QString& pref, QDate val );

	/**
	 * Set the specified preference @p pref to the given value @p val.
	 * @param pref The preference in question
	 * @param val The new preference value
	 */
	void setWP( const QString& pref, QFont val );

	/**
	 * Set the specified preference @p pref to the given value @p val.
	 * @param pref The preference in question
	 * @param val The new preference value
	 */
	void setFP( const QString& pref, float val );

	/**
	 * Set the specified preference @p pref to the given value @p val.
	 * @param pref The preference in question
	 * @param val The new preference value
	 */
	void setBP( const QString& pref, bool val );

	/**
	 * Set the specified preference @p pref to the given value @p val.
	 * @param pref The preference in question
	 * @param val The new preference value
	 */
	void setIP( const QString& pref, int val );

	/**
	 * Get the AccountTypes table data
	 * @param rs A dynamically-created ResultSet with the results of the query
	 * (this has to be deleted by the calling object.)
	 */
	void getATs( QHaccResultSet *& rs );

	/**
	 * Write the given transaction to the file @p file.
	 * @param tid The ID of the transaction to be written
	 * @param file The filename to written
	 * @param err A description of the errors thrown, if any.
	 * @return True, if the writing was successful.
	 */
	bool writeTransToFile( uint tid, const QString& file, QString& err );

	/**
	 * Take a whole-number value of money and create a string of that value,
	 * including decimals.
	 * @param cents The integer value of money
	 * @return A string representing the integer as fractional money.
	 */
	static QString convMoney( int cents );

	/**
	 * Take a string with decimal places and convert it to an integer
	 * representation.
	 * @param str The string containing decimals
	 * @return The integer representation of the money.
	 */
	static int convMoney( const QString& str );

	/**
	 * Get the directory containing i18n data from the preconf file.
	 * @return The directory containing i18n data.
	 */
	QString languagedir() const;
	
	/**
	 * Get the descriptions of available plugins of a given type
	 * @param type The type of plugins to return
	 * @param stubs A dynamically-allocated array of plugin identifiers
	 * (This must be deleted by the caller.)
	 * @param descriptions A dynamically-allocated array of plugin descriptions
	 * (This must be deleted by the caller.)
	 * @param filenames A dynamically-allocated array of plugin files
	 * (This must be deleted by the caller.)
	 * @param arrsize The size of the newly-created @p descriptions and 
	 * @p filenames.
	 * @return The index of @p filenames containing the currently-used
	 * plugin, or -1 if none.
	 */
	int getPluginInfo( int type, QString *& stubs, QString *& descriptions,
										 QString *& filenames, int& arrsize ) const;

	/**
	 * Given a home, pass a db plugin capable of reading that datastore
	 * @param type The type of plugin desired
	 * @param home The QHACC_HOME that needs to be read
	 * @param pi The QHaccPlugin that can read the given home
	 * (This must be deleted by the caller using destroyPlugin() )
	 */
	QString getPluginFor( int type, const QString& home,
												QHaccPlugin *& pi ) const;
	
	/**
	 * Destroy the given plugin. This requires the correct library to 
	 * be reloaded and the right destructor to be called from that library
	 * @param type The type of plugin we're dealing with
	 * @param pi The plugin to destroy
	 * @return True if the plugin was destroyed
	 */
	bool destroyPlugin( int type, QHaccPlugin * pi );

	/**
	 * Recalculate the balance, reconciled balance, and budget values of the
	 * given account.
	 * @param a The account to recalculate
	 * @return A copy of the new information in the Accounts table
	 */
	Account calcBalOfA( const Account& );

	/**
	 * Get the transactions whose parent is the given account, and all
	 * transactions related to them by splitgroup.
	 * @param a The account in question
	 * @param rs A dynamically-created ResultSet with the results of the query
	 * (this has to be deleted by the calling object.)
	 */
	void getTForArchive( const Account & a, QHaccResultSet *& rs );

	/**
	 * Recalculate all account balances, and if different than expected, reset
	 * the account's opening balance to keep the current balance constant.
	 */
	void resetOBals();

	/**
	 * Make a creation script for creating the given QHACC_HOME
	 * @param str The QHACC_HOME to create
	 * @return A string containing all the commands necessary for creating 
	 * a QHACC_HOME
	 */
	QString create( const QString& str );

	/**
	 * Query the database
	 * @param table The table to query
	 * @param ts An array of TableSelect objects
	 * @param tssz The size of the ts array
	 * @param rs A dynamically-created ResultSet with the results of the query
	 * (this has to be deleted by the calling object.)
	 * @return True if the result set is not empty
	 */
	bool getWhere( Table table, TableSelect * ts, uint tssz,
								 QHaccResultSet *& rs);

	/**
	 * Query the database
	 * @param table The table to query
	 * @param tg columns to retrieve
	 * @param ts An array of TableSelect objects
	 * @param tssz The size of the ts array
	 * @param rs A dynamically-created ResultSet with the results of the query
	 * (this has to be deleted by the calling object.)
	 * @return True if the result set is not empty
	 */
	bool getWhere( Table table, const TableGet& tg, TableSelect * ts, uint tssz,
								 QHaccResultSet *& rs);

	/**
	 * @return The number of rows in the given table
	 */
	uint cnt( Table );

	/**
	 * @param t The table on which to operate
	 * @param col The column in question
	 * @returns the minimum value of the column
	 */
	TableCol min( Table t, int col );

	/**
	 * Remove extraneous transactions from the given table. This condenses
	 * transactions with duplicate accounts into one transaction.
	 * @param t A set of transactions
	 * @return True, if the newly-condensed @p t is valid for insertion.
	 */
	bool condenseSG( QHaccTable * t ) const;

 protected:
	/**
	 * Delete from the given table every row where the given column matches the
	 * given selection criteria. This function allows direct database access, so
	 * it should be used with extreme care. 
	 * @param t The table on which to operate
	 * @param ts The selection criteria
	 * @return the number of rows affected
	 */
	uint deleteWhere( Table t, const TableSelect& ts );

	/**
	 * Delete the given rows from the given table. This function allows direct
	 * database access, so it should be used with extreme care. 
	 * @param t The table on which to operate
	 * @param rs The rows which need to be deleted
	 * @return the number of rows affected
	 */
	uint deleteWhere( Table t, QHaccResultSet * rs );

	/**
	 * Update the given table in the given way. This function allows direct 
	 * database access, so it should be used with extreme care. 
	 * @param t The table on which to operate
	 * @param ts The selection criteria
	 * @param tu What to update in the returned rows
	 * @return number of rows affected
	 */
	uint updateWhere( Table t, const TableSelect& ts, const TableUpdate& tu );

	/**
	 * Load the given rows into the given table. This function allows direct
	 * database access, so it should be used with extreme care.
	 * @param t The table on which to operate
	 * @param rs The rows to be added
	 * @return True, if the rows are added
	 */
	bool load( Table t, QHaccResultSet& rs );

	/**
	 * @param t The table on which to operate
	 * @param col The column in question
	 * @returns the maximum value of the column
	 */
	TableCol max( Table t, int col );

 private:
	QString qhacchome, langdir;
	QHaccDBPlugin * db;
	PluginManager ** plugmen;

	/**
	 * All preferences are kept as strings in the database. This function
	 * is called by all the other setP functions to actually maintain the
	 * database state.
	 */
	bool isetP( const QString& pref, QString val );

	/**
	 * Search the given directory for plugins and load them. Also, set any
	 * preconfiguration options specified in the preconf file.
	 * @param rootdir The top of the searchable directory structure.
	 */
	void readpre( const QString& rootdir );

 signals:
	/**
	 * Added the given account
	 */
	void addedA( const Account& );

	/**
	 * Removed the given account
	 */
	void removedA( const Account& );

	/**
	 * Updated the account @p old to look like @p newer
	 * @param old The old account
	 * @param newer The new account	 
	 */
	void updatedA( const Account& old, const Account& newer );

	/**
	 * The given account is out of balance.
	 * @param a The account in question
	 * @param bal The new balance of the account
	 */
	void overBudget( const Account& a, int bal );

	/**
	 * Added the given transaction
	 */
	void addedT( const Transaction& );

	/**
	 * Removed the given transaction
	 */
	void removedT( const Transaction& );

	/**
	 * updated the given transaction
	 */
	void updatedT( const Transaction& );

	/**
	 * Added any transaction
	 */
	void addedT();

	/**
	 * Removed any transaction
	 */
	void removedT();

	/**
	 * Updated any transaction
	 */
	void updatedT();

	/**
	 * Added the given ledger
	 */
	void addedL( const Ledger& );

	/**
	 * Removed the given ledger
	 */
	void removedL( const Ledger& );

	/**
	 * Updated the ledger @p old to look like @p newer
	 * @param old The old ledger
	 * @param newer The new ledger
	 */
	void updatedL( const Ledger& old, const Ledger& newer );

	/**
	 * Changed the given preference
	 */
	void changedP( const QString&, QString );

	/**
	 * Changed the given preference
	 */
	void changedP( const QString&, QColor );

	/**
	 * Changed the given preference
	 */
	void changedP( const QString&, QDate );

	/**
	 * Changed the given preference
	 */
	void changedP( const QString&, QFont );

	/**
	 * Changed the given preference
	 */
	void changedP( const QString&, float );

	/**
	 * Changed the given preference
	 */
	void changedP( const QString&, bool );

	/**
	 * Changed the given preference
	 */
	void changedP( const QString&, int );

	/**
	 * The database is in a dirty state
	 * @param b True, if the database should be saved
	 */
	void needSave( bool b );
};


class PluginManager {
 public:
	PluginManager( const QString& root, const QString& extension );
	virtual ~PluginManager();
	
	virtual QString getPluginFor( const QString& key, QHaccPlugin *& pi ) const;
	virtual bool destroyPlugin( QHaccPlugin * pi );
	virtual void getPluginInfo( QString *& stubs, QString *& descrs, QString *& filenames,
															int& sz ) const;

 private:
	int numPI, * pluginUsage;
	QString * pluginDescs, * pluginFiles, * pluginKnows;
	QLibrary ** libs;
};

#endif
